/*
class describing IGC/KML tracklog

Author: Daniel Dimov <danieldimov@gmail.com>
*/

package org.fai.fs;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.ArrayList;
import com.ibm.util.CoordinateConversion;
import org.joda.time.*;
import org.joda.time.format.*;

public class Tracklog {
	final static Charset ENCODING = StandardCharsets.ISO_8859_1;
	private int startTime, endTime, fixNum, offset;
	private boolean geoid, isComplete, hasGpsAlt, hasBaroAlt;
	private String header;
	private Path source;
	private int[] time, baroAlt, gpsAlt, trueAlt, place, eastning, northning; 
	private float[] lat, lon, pressure;
	private DateTime tracklogDate;
	
	public Tracklog(Path sourceFile, int st) throws IOException {
		fixNum = 0;
		geoid = false;
		isComplete = false;
		hasGpsAlt = false;
		hasBaroAlt = false;
		offset = 0;
		header = "";
		
		if (sourceFile.toString().toLowerCase().endsWith(".igc")) { // load IGC file
			ArrayList<String> lines = new ArrayList<String>();
			int i = 0;
			DateTimeFormatter fmt = DateTimeFormat.forPattern("ddMMyy");
			DateTime maxdate = fmt.parseDateTime("010100");
			String line = null;
			try (BufferedReader reader = Files.newBufferedReader(sourceFile, ENCODING)){
				while ((line = reader.readLine()) != null) {
					lines.add(line);
					if (line.startsWith("B")) {
						i++; // count B-records
					} else if (line.startsWith("HFDTE")) {
						DateTime cd = fmt.parseDateTime(line.substring(5,11));
						if (cd.isAfter(maxdate)) maxdate = cd; // find the latest date found in the IGC file
					}
				}
			}
			
			time = new int [i];
			baroAlt = new int [i];
			gpsAlt = new int [i];
			trueAlt = new int [i];
			place = new int [i];
			eastning = new int [i];
			northning = new int [i];
			lat = new float [i];
			lon = new float [i];
			pressure = new float [i];
			
			int j = 0;
			int t = 0;
			int lt = 0;
			boolean readFlag = false;
			CoordinateConversion cc = new CoordinateConversion();
			for (String l : lines) {
				if (l.startsWith("A") || l.startsWith("H")) {
					header = header + l + System.lineSeparator();
					if (l.startsWith("HFDTE")) {
						DateTime cd = fmt.parseDateTime(l.substring(5,11));
						if (cd.compareTo(maxdate) == 0) readFlag = true; // start reading only after the latest date in the IGC file is reached
					}
				} else if (l.startsWith("B") && readFlag) {
					t = Integer.parseInt(l.substring(1,3)) * 3600 + Integer.parseInt(l.substring(3,5)) * 60 + Integer.parseInt(l.substring(5,7));
					if (t >= st) {
						if (t != lt) { // parse only fixes with unique time
							time[j] = t;
							lat[j] = ((float) Integer.parseInt(l.substring(7,9)) + 
								((float) Integer.parseInt(l.substring(9,14)))/60000f ) * 
								( l.substring(14,15).equals("N") ? 1f : -1f );
							lon[j] = ((float) Integer.parseInt(l.substring(15,18)) + 
								((float) Integer.parseInt(l.substring(18,23)))/60000f ) * 
								( l.substring(23,24).equals("E") ? 1f : -1f );
							if (lat[j] < 80f && lat[j] > -80f && lon[j] < 180f && lon[j] > -180f) {
								baroAlt[j] = Integer.parseInt(l.substring(25,30));
								if (baroAlt[j] !=0) hasBaroAlt = true;
								gpsAlt[j] = Integer.parseInt(l.substring(30,35));
								if (gpsAlt[j] !=0) hasGpsAlt = true;
								pressure[j] = (1013.25f * (float) Math.pow( 1f - 0.0065f * baroAlt[j] / 288.15f , 5.256f));
								trueAlt[j]=0;
								String s = cc.latLon2UTM(lat[j], lon[j]);
								String[] utm = s.split(" ");
								eastning[j] = Integer.parseInt(utm[2]);
								northning[j] = Integer.parseInt(utm[3]);
								j++;
								lt = t;
							}
						}
					}
				}
			}
			
			if (j > 0) {
				startTime = time[0];
				endTime = time[j - 1];
			}
			fixNum = j;
			isComplete = hasBaroAlt && hasGpsAlt && fixNum > 0;
			source = sourceFile;
			tracklogDate = maxdate;
			
			lines = null;

			geoid = false;			
			/*********************************************************************
			Here must be the code setting properly the geoid flag of the object.
			By default in IGC files the altitude MUST BE above the ellipsoid
			and therefore geoid flag must be false.
			In reality there are instruments producing IGC files with altitude
			above geoid. If for some instruments this fact is known - they must
			be detected (by the header) here and the flag must be set to true!
			*********************************************************************/
			if (header.contains("Geoid")) geoid = true;
			
			
		// here load IGC file ends
		} else if (sourceFile.toString().toLowerCase().endsWith(".kml")) { // load KML file
			
			/* IMPLEMENT KML LOAD HERE */
			
		// here load KML file ends
		} else { // other file
		
		};
	} // end of the constructor
	
	public void save(Path saveFile, boolean truealt, boolean expcodes) throws IOException {
	
		if (saveFile.toString().toLowerCase().endsWith(".igc")) { // save IGC file
			try (BufferedWriter writer = Files.newBufferedWriter(saveFile, ENCODING)){
				writer.write(header);
				writer.write((expcodes ? "HFXF1ALTTYPE:" : "HFALTTYPE:") + (geoid ? "Geoid" : "Ellipsoid"));
				writer.newLine();
				if (truealt) {
					if (expcodes) {
						writer.write("I033640XF24145XF34651XF4");
					} else {
						writer.write("I033640ABR4145AGN4651APR");
					}
				} else {
					if (expcodes) {
						writer.write("I023640XF54146XF4");
					} else {
						writer.write("I023640ATR4146APR");
					}
				}
				writer.newLine();
				
				String l;
				float la, lo;
				int lad, lod;
				for(int i=0; i < fixNum; i++) {
					l = "B" + String.format("%02d", time[i] / 3600) + String.format("%02d", (time[i] % 3600) / 60) + String.format("%02d", time[i] % 60);
					la = Math.abs(lat[i]); lad = Math.round((la - (int) la)*60000f);
					lo = Math.abs(lon[i]); lod = Math.round((lo - (int) lo)*60000f);
					l = l + String.format("%02d", (int) la) + String.format("%05d", lad) + (lat[i] >= 0 ? "N" : "S");
					l = l + String.format("%03d", (int) lo) + String.format("%05d", lod) + (lon[i] >= 0 ? "E" : "W");
					if (truealt) {
						l = l + "A" + String.format("%05d", trueAlt[i] + offset) + String.format("%05d", trueAlt[i] + offset);
						l = l + String.format("%05d", baroAlt[i]) + String.format("%05d", gpsAlt[i]) + String.format("%06d", Math.round(pressure[i] * 100));
					} else {
						l = l + "A" + String.format("%05d", baroAlt[i]) + String.format("%05d", gpsAlt[i]);
						l = l + String.format("%05d", trueAlt[i] + offset) + String.format("%06d", Math.round(pressure[i] * 100));
					}
					writer.write(l);
					writer.newLine();
				}
			}
		// here save IGC file ends
		} else if (saveFile.toString().toLowerCase().endsWith(".kml")) { // save KML file
			
			/* IMPLEMENT KML SAVE HERE */
			
		// here save KML file ends
		} else { // other file
		
		};
	
	} // save method
	
	public String getHeader() {
		return header;
	}
	
	public DateTime getDate() {
		return tracklogDate;
	}
	
	public Path getSource() {
		return source;
	}

	public int getStartTime() {
		return startTime;
	}
	
	public int getEndTime() {
		return endTime;
	}
	
	public int getFixNum() {
		return fixNum;
	}
		
	public boolean getGeoid() {
		return geoid;
	}
	
	public void setGeoid(boolean g) {
		geoid = g;
	}
	
	public int getTime(int i) {
		if (fixNum > 0) {
			return time[i];
		} else {
			return 0;
		}
	}

	public float getLat(int i) {
		if (fixNum > 0) {
			return lat[i];
		} else {
			return 0f;
		}
	}

	public float getLon(int i) {
		if (fixNum > 0) {
			return lon[i];
		} else {
			return 0f;
		}
	}
	
	public int getEastning(int i) {
		if (fixNum > 0) {
			return eastning[i];
		} else {
			return 0;
		}
	}

	public int getNorthning(int i) {
		if (fixNum > 0) {
			return northning[i];
		} else {
			return 0;
		}
	}
	
	public float getPressure(int i) {
		if (fixNum > 0) {
			return pressure[i];
		} else {
			return 0f;
		}
	}

	public int getBaroAlt(int i) {
		if (fixNum > 0) {
			return baroAlt[i];
		} else {
			return 0;
		}
	}

	public int getAvgBaroAlt() {
		if (fixNum > 0) {
			int s = 0;
			for (int i = 0; i < fixNum; i++) s += baroAlt[i];
			return s / fixNum;
		} else {
			return 0;
		}
	}

	public int getGpsAlt(int i) {
		if (fixNum > 0) {
			return gpsAlt[i];
		} else {
			return 0;
		}
	}

	public int getAvgGpsAlt() {
		if (fixNum > 0) {
			int s = 0;
			for (int i = 0; i < fixNum; i++) s += gpsAlt[i];
			return s / fixNum;
		} else {
			return 0;
		}
	}

	public int getTrueAlt(int i) {
		if (fixNum > 0) {
			return trueAlt[i];
		} else {
			return 0;
		}
	}

	public int getAvgTrueAlt() {
		if (fixNum > 0) {
			int s = 0;
			for (int i = 0; i < fixNum; i++) s += trueAlt[i];
			return s / fixNum;
		} else {
			return 0;
		}
	}

	public void setTrueAlt(int i, int a) {
		if (fixNum > 0 && i < fixNum) {
			trueAlt[i] = a;
		}
	}
	
	public int getPlace(int i) {
		if (fixNum > 0) {
			return place[i];
		} else {
			return 0;
		}
	}

	public void setPlace(int i, int p) {
		if (fixNum > 0 && i < fixNum) {
			place[i] = p;
		}
	}
	
	public int getOffset() {
		return offset;
	}

	public void setOffset(int o) {
		offset = o;
	}
	
	public boolean isEmpty() {
		return (fixNum == 0);
	}
	
	public boolean isComplete() {
		return isComplete;
	}
	
	// this method returns the index with tolerance dt from time t 
	public int getIndex(int t, int dt) {
		int rl, rh, ci, ct;
		
		if ( ! this.isEmpty() && t >= startTime && t <= endTime ) {
			rl = 0;
			rh = fixNum - 1;
			if (Math.abs(t - rl) <= dt) { return rl; }
			if (Math.abs(t - rh) <= dt) { return rh; }
			//ci = (int) ((float)(rh - rl) * (float)(t - time[rl]) / (float)(time[rh] - time[rl])) + rl;
			ci = (int) ((float)(rh - rl) / 2) + rl;
			ct = time[ci];
			while (Math.abs(ct - t) > dt && rh - rl > 2) {
				if (ct > t) {
					rh = ci;
				} else {
					rl = ci;
				}
				ci = (int) ((float)(rh - rl) / 2) + rl;
				ct = time[ci];
			}
			if (Math.abs(t - ct) <= dt) {
				return ci;
			} else {
				return -1;
			}
		} else {
			return -1;
		}
	} // getIndex method
	
} // Tracklog class