/*
main class of the TracklogAnalizer program

Author: Daniel Dimov <danieldimov@gmail.com>
*/

package org.fai.fs;

import java.io.IOException;
import java.io.File;
import java.io.FilenameFilter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Scanner;
import java.util.Random;
import java.text.NumberFormat;
import org.fai.fs.Tracklog;
import org.fai.fs.MyPlot;
import flanagan.analysis.Regression;
import flanagan.analysis.CurveSmooth;
import org.joda.time.*;
import org.joda.time.format.*;



public class TracklogAnalizer {
	/*
	// constants based on dry adiabatic laps rate
	final static double ATMC = 3.485977388d;
	final static double ATMIC = 0.286863593d;
	final static double LAPSR = 0.0098d;
	*/
	// constants based on average laps rate
	final static double ATMC = 5.255781293d;
	final static double ATMIC = 0.190266669d;
	final static double LAPSR = 0.0065d;

    public static void main(String[] args) throws IOException {
		int maxplace = 0;
		int maxtime = 0;
		int mintime = 24*60*60;
		double[] cb, ct;
		double sb, st;
		int identtr1 = -1;
		int identtr2 = -1;
		int nid = 0;
		int geoidH = 0;
		int startTime = -1;
		boolean noSave = false;
		boolean saveExperimental = false;
		boolean noTrueAltitude = false;
		
		
		/* basic calculations for b0 and T0
		double p1, p2, b, t, c, ic, l, lat, lon;
		int hb1, hb2, hg1, hg2, ht1, ht2;
		
		c = 5.255781293d; ic = 0.190266669d; l = 0.0065d;
		hg1 = 166; hb1 = 200;
		hg2 = 1492; hb2 = 1470;
		lat = 43d + 13.255d / 60d; lon = 27d + 55.101d / 60d;
		
		CoordinateConversion cc = new CoordinateConversion();
		String utm = cc.latLon2UTM(lat, lon);
		System.out.println("UTM coordinates: " + utm);
		
		p1 = 1013.25d * Math.pow( 1d - l * hb1 / 288.15d , c);
		p2 = 1013.25d * Math.pow( 1d - l * hb2 / 288.15d , c);
		System.out.println("p1 = " + p1);
		System.out.println("p2 = " + p2);
		
		b = Math.pow( (hg2 * Math.pow(p1, ic) - hg1 * Math.pow(p2, ic))/(hg2 - hg1), c);
		System.out.println("b = " + b);
		
		t = hg1 * l / ( 1d - Math.pow(p1/b, ic)) - 273.15d;
		System.out.println("t1 = " + t);
		t = hg2 * l / ( 1d - Math.pow(p2/b, ic)) - 273.15d;
		System.out.println("t2 = " + t);
		
		
		ht1 = (int) Math.round(((273.15d + t)/l) * ( 1d - Math.pow(p1/b, ic)));
		ht2 = (int) Math.round(((273.15d + t)/l) * ( 1d - Math.pow(p2/b, ic)));
		System.out.println("ht1 = " + ht1);
		System.out.println("ht2 = " + ht2);
		

		float la, lo;
		String l = "B1529310428577N07609190WA016150174200000";
		la = ((float) Integer.parseInt(l.substring(7,9)) + 
			(((float) Integer.parseInt(l.substring(9,14)))/1000f)/60f ) * 
			( l.substring(14,15).equals("N") ? 1f : -1f );
		lo = ((float) Integer.parseInt(l.substring(15,18)) + 
			((float) Integer.parseInt(l.substring(18,23))/1000f)/60f ) * 
			( l.substring(23,24).equals("E") ? 1f : -1f );
		
		System.out.println("lat=" + la + " " + l.substring(14,15));
		System.out.println("lon=" + lo + " " + l.substring(23,24));
		System.out.println( ((float) 28577)/60000f );
		*/
		
		if (args.length == 0) {
			System.out.println("Usage: java TracklogAnalizer [-gh <geoid height>] [-st hh:mm] [-ns | -s] [-exp | -fin] [-nta | -ta] <source dir>");
			System.exit(0);
		} else {
			for (int i = 0; i < args.length; i++) {
				switch (args[i]) {
					case "-ns": noSave = true; break;
					case "-s": noSave = false; break;
					case "-exp": saveExperimental = true; break;
					case "-fin": saveExperimental = false; break;
					case "-nta": noTrueAltitude = true; break;
					case "-ta": noTrueAltitude = false; break;
					case "-gh": geoidH = Integer.parseInt(args[i+1]); break;
					case "-st": DateTimeFormatter fmt = DateTimeFormat.forPattern("HH:mm");
								DateTime ststr = fmt.parseDateTime(args[i+1]);
								startTime = ststr.getHourOfDay() * 3600 + ststr.getMinuteOfHour() * 60;
								//System.out.println("Start time: " + startTime);
								break;
				}
			}
		}
		
		
		System.out.println("Reading IGC files in directory " + args[args.length - 1]);
		Path path = Paths.get(args[args.length - 1]);
		FilenameFilter filefilter = new FilenameFilter() { public boolean accept(File dir, String name) { return name.toLowerCase().endsWith(".igc"); } };
		File [] files = path.toFile().listFiles(filefilter);
		ArrayList<Tracklog> tracks = new ArrayList<Tracklog>();
		for (int i = 0; i < files.length; i++){
			if (files[i].isFile()) {
				// System.out.println("Loading file " + files[i]);
				tracks.add(new Tracklog(files[i].toPath(), startTime));
				// System.out.println("File " + tracks.get(tracks.size()-1).getSource().toString() + " has date " + tracks.get(tracks.size()-1).getDate().toString());
			}
		}
		System.out.println("IGC files read: " + tracks.size());
		
		// check if any of the tracklogs has different date
		DateTime cd = tracks.get(0).getDate();
		for (int i = 1; i < tracks.size(); i++) {
			if ( ! tracks.get(i).isComplete() ) System.out.println("Tracklog file " + tracks.get(i).getSource() + " is NOT complete!");
			if ( cd.compareTo(tracks.get(i).getDate()) != 0 ) {
				System.out.println("Tracklog file " + tracks.get(i).getSource() + " has different date!");
				System.out.println(cd.toString());
				System.out.println(tracks.get(i).getDate().toString());
				System.exit(1);
			}
		}
		
		/* time analysis of the tracklogs
		System.out.println("Searching for 1000 random times");
		Random randomGenerator = new Random();
		int[] s = new int[10000];
		int[] f = new int[10000];
		int[] tr = new int[10000];
		int rt, rs, j;
		for (int i = 0; i < 10000; i++) {
			rt = randomGenerator.nextInt(tracks.size());
			rs = randomGenerator.nextInt(tracks.get(rt).getEndTime() - tracks.get(rt).getStartTime()) + tracks.get(rt).getStartTime();
			s[i] = rs;
			j = tracks.get(rt).getIndex(rs, 3);
			f[i] = tracks.get(rt).getTime(j);
			tr[i] = rt;
		}

		for (int t = 0; t < 60; t++) {
			int n = 0;
			for (int i = 0; i < 10000; i++) {
				if(Math.abs(s[i] - f[i]) == t) n++;
			}
			System.out.println("Number of times with difference " + t + " is " + n);
		}
		System.out.println("Done");
		*/
				
		{ // this block deals with place approximation
			System.out.println("Preparing place approximation with 10000 random tracklog points");
			Random randomGenerator = new Random();
			double[] x = new double[10000];
			double[] y = new double[10000];
			int rt, j;
			double xmin = 0;
			double xmax = 0;
			for (int i = 0; i < 10000; i++) {
				rt = randomGenerator.nextInt(tracks.size());
				if (tracks.get(rt).getFixNum() > 0) {
					j = randomGenerator.nextInt(tracks.get(rt).getFixNum());
					x[i] = (double) tracks.get(rt).getEastning(j);
					y[i] = (double) tracks.get(rt).getNorthning(j);
					if (i == 0) { xmin = x[i]; xmax = x[i]; }
					if (x[i] < xmin) { xmin = x[i]; }
					if (x[i] > xmax) { xmax = x[i]; }
				}
			}
			Regression regr = new Regression(x, y);
			regr.linear();
			double[] c = regr.getBestEstimates();
			for (int i = 0; i < c.length; i++) {
				System.out.println("C[" + i + "] = " + c[i]);
			}
			double b = c[0];
			double a = c[1];
			double x0 = (xmin + xmax)/2d;
			double y0 = a * x0 + b;
			double ymax = a * xmax + b;
			double L = Math.sqrt( (xmax - x0)*(xmax - x0) + (ymax - y0)*(ymax - y0) );
			regr = null;
			x = null;
			y = null;
						
			System.out.println("Storing places into tracklogs");
			int maxptr = 0;
			for (int tr = 0; tr < tracks.size(); tr++) {
				boolean zeroBAFlag = false;
				for (int i = 0; i < tracks.get(tr).getFixNum(); i++) {
					double xc = (double) tracks.get(tr).getEastning(i);
					double yc = (double) tracks.get(tr).getNorthning(i);
					int p = (int) (((y0 - yc)*(y0 - ymax) - (x0 - xc)*(xmax - x0))/L);
					tracks.get(tr).setPlace(i, p);
					if ( tracks.get(tr).getTime(i) > maxtime ) maxtime = tracks.get(tr).getTime(i);
					if ( tracks.get(tr).getTime(i) < mintime ) mintime = tracks.get(tr).getTime(i);
					if ( Math.abs(p) > maxplace ) { maxplace = Math.abs(p); maxptr = tr; }
					if ( tracks.get(tr).getBaroAlt(i) == 0 ) zeroBAFlag = true;
				}
				if (zeroBAFlag && tracks.get(tr).isComplete()) System.out.println("Log with ZERO baro-altitude forund: " + tracks.get(tr).getSource());
			}
			System.out.println("Storing places done");
			System.out.println("Maximum absolute value of place: " + maxplace + " in " + tracks.get(maxptr).getSource());
		}

		{ // this block finds two identical tracklogs
			System.out.println("Calculating average distances between tracklogs");
			long[][] ds = new long[tracks.size()][tracks.size()];
			long maxds = 0;
			long minds = -1;
			int mindstr1 = -1;
			int mindstr2 = -1;
			int mindpn = 0;
			nid = 0;
			for (int tr1 = 0; tr1 < tracks.size(); tr1++) {
				for (int tr2 = tr1; tr2 < tracks.size(); tr2++) {
					if (tr1 != tr2 && tracks.get(tr1).isComplete() & tracks.get(tr2).isComplete()) {
						long distsum = 0;
						int pn = 0;
						for (int time = mintime; time < maxtime; time += 10) {
							int i1 = tracks.get(tr1).getIndex(time, 3);
							int i2 = tracks.get(tr2).getIndex(time, 3);
							if (i1 >= 0 && i2 >= 0) {
								int e1 = tracks.get(tr1).getEastning(i1);
								int n1 = tracks.get(tr1).getNorthning(i1);
								int e2 = tracks.get(tr2).getEastning(i2);
								int n2 = tracks.get(tr2).getNorthning(i2);
								distsum += (long) Math.sqrt( Math.pow((double) (e1 - e2), 2d) + Math.pow((double) (n1 - n2), 2d) );
								pn++;
							}
						}
						if (pn > 0) {
							distsum /= pn;
							if (distsum < 50) nid++;
							ds[tr1][tr2] = distsum;
							if (minds == -1) {
								minds = distsum;
								mindstr1 = tr1;
								mindstr2 = tr2;
								mindpn = pn;
							}
							if (minds > distsum) {
								minds = distsum;
								mindstr1 = tr1;
								mindstr2 = tr2;
								mindpn = pn;
							}
						}
					}
				}
			}
			
			if (minds != -1) {
				System.out.println("Minimum average distance found: " + minds + "m between tracks " + tracks.get(mindstr1).getSource() + " and " + tracks.get(mindstr2).getSource());
			}

			if (nid == 1) { // one identical pair
				System.out.println("Found identical tracklogs. Vertical error analysis will be performed.");
				identtr1 = mindstr1;
				identtr2 = mindstr2;
			} else if (nid > 1) { // more identical pairs
				System.out.println("More than one identical pairs are found.");
				minds = -1;
			} else if (nid == 0) { // no identical pairs
				System.out.println("No identical tracklogs.");
				minds = -1;
			} else {
				System.exit(1);
			}
			//System.exit(0);
		} // end of the block finding two identical tracklogs

		
		{ // this block calculates the best-fit coeficents for b and t
			double[] px = new double[10000];
			double[] pb = new double[10000];
			double[] pt = new double[10000];
			double[] pp1 = new double[10000];
			double[] ph1 = new double[10000];
			double[] pp2 = new double[10000];
			double[] ph2 = new double[10000];
			
			Random randomGenerator = new Random();
			int n = 0;
			sb = 0;
			st = 0;
			
			System.out.println("mintime = " + mintime + ", maxtime = " + maxtime);
			for (int time = mintime; time < maxtime; time += 5) {
				//System.out.println("Trying to find 100 good points for " + per + "% of the total task time");
				//int time = (maxtime - mintime)*per/100 + mintime;
				int maxi = -1;
				int maxh = 0;
				int maxtr = -1;
				int mini = -1;
				int minh = 0;
				int mintr = -1;
				
				for (int tr1 = 0; tr1 < tracks.size(); tr1++) {
					for (int tr2 = tr1; tr2 < tracks.size(); tr2++) {
						if (tr1 != tr2 && tracks.get(tr1).isComplete() & tracks.get(tr2).isComplete()) {
							int i1 = tracks.get(tr1).getIndex(time, 4);
							int i2 = tracks.get(tr2).getIndex(time, 4);
							if (i1 >= 0 && i2 >= 0) {
								int h1 = tracks.get(tr1).getGpsAlt(i1) - (tracks.get(tr1).getGeoid() ? 0 : geoidH);
								int h2 = tracks.get(tr2).getGpsAlt(i2) - (tracks.get(tr2).getGeoid() ? 0 : geoidH);
								int pl1 = tracks.get(tr1).getPlace(i1);
								int pl2 = tracks.get(tr2).getPlace(i2);
								if (n < 10000 && Math.abs(h1 - h2) > 600 && h1 > 0 && h2 > 0 && Math.abs(pl1 - pl2) < 5000) { 
									/*  this code is for adding all good points to the preliminary arrays
									float p1 = tracks.get(tr1).getPressure(i1);
									float p2 = tracks.get(tr2).getPressure(i2);
									pb[n] = Math.pow( (h2 * Math.pow(p1, ATMIC) - h1 * Math.pow(p2, ATMIC))/(h2 - h1), ATMC);
									pt[n] = h1 * LAPSR / ( 1d - Math.pow(p1/pb[n], ATMIC));
									//px[n] = (((double) (pl1 + pl2))/2d) / ((double) maxplace);
									px[n] = ((double)(time - mintime))/((double)(maxtime - mintime));
									if (pb[n] > 500 && pb[n] < 1500 && pt[n] > 0 && pt[n] < 600) n++;
									*/
									if (maxtr == -1) {
										maxi = (h1 > h2 ? i1 : i2);
										maxtr = (h1 > h2 ? tr1 : tr2);
										maxh = (h1 > h2 ? h1 : h2);
									} else {
										if (h1 > maxh || h2 > maxh) {
											maxi = (h1 > h2 ? i1 : i2);
											maxtr = (h1 > h2 ? tr1 : tr2);
											maxh = (h1 > h2 ? h1 : h2);
										}
									}
									if (mintr == -1) {
										mini = (h1 < h2 ? i1 : i2);
										mintr = (h1 < h2 ? tr1 : tr2);
										minh = (h1 < h2 ? h1 : h2);
									} else {
										if (h1 < minh || h2 < minh) {
											mini = (h1 < h2 ? i1 : i2);
											mintr = (h1 < h2 ? tr1 : tr2);
											minh = (h1 < h2 ? h1 : h2);
										}
									}
								}
							}
						}
					}
				}

				if (maxtr != -1 && mintr != -1) {
					double p1 = tracks.get(maxtr).getPressure(maxi);
					double p2 = tracks.get(mintr).getPressure(mini);
					int h1 = maxh;
					int h2 = minh;
					pb[n] = Math.pow( (h2 * Math.pow(p1, ATMIC) - h1 * Math.pow(p2, ATMIC))/(h2 - h1), ATMC);
					pt[n] = h1 * LAPSR / ( 1d - Math.pow(p1/pb[n], ATMIC));
					pp1[n] = p1;
					pp2[n] = p2;
					ph1[n] = h1;
					ph2[n] = h2;
					
					// this is for the b, t dependence of time
					px[n] = ((double)(time - mintime))/((double)(maxtime - mintime));
					
					// this is for the b, t dependence of place
					//px[n] = (double) (tracks.get(maxtr).getPlace(maxi) + tracks.get(mintr).getPlace(mini)) / 2d / (double) maxplace;
					
					// this is check for possible in reality values of b and t; b(900, 1100) hPa; t(-73, 123) deg C
					if (pb[n] > 900 && pb[n] < 1100 && pt[n] > 200 && pt[n] < 400) { 
						sb += pb[n];
						st += pt[n];
						n++;
					}
				}
				
			}
			
			if (n >= 100) {
				System.out.println("Found " + n + " good points of the total task time");
				double[] x = new double[n];
				double[] b = new double[n];
				double[] t = new double[n];
				double[] p1 = new double[n];
				double[] p2 = new double[n];
				double[] h1 = new double[n];
				double[] h2 = new double[n];
				System.arraycopy(px, 0, x, 0, n);
				System.arraycopy(pb, 0, b, 0, n);
				System.arraycopy(pt, 0, t, 0, n);
				System.arraycopy(pp1, 0, p1, 0, n);
				System.arraycopy(pp2, 0, p2, 0, n);
				System.arraycopy(ph1, 0, h1, 0, n);
				System.arraycopy(ph2, 0, h2, 0, n);
				
				Regression rb = new Regression(x, b);
				rb.linear();
				cb = rb.getBestEstimates();
				for (int j = 0; j < cb.length; j++) {
					System.out.println("b: c[" + j + "] = " + cb[j]);
				}
				//rb.plotXY();
				rb = null;

				Regression rt = new Regression(x, t);
				rt.linear();
				ct = rt.getBestEstimates();
				for (int j = 0; j < ct.length; j++) {
					System.out.println("t: c[" + j + "] = " + ct[j]);
				}
				//rt.plotXY();
				rt = null;
				
				{ // GRAL plotting
					//MyPlot plot = new MyPlot("Good points", p1, h1, p2, h2);
					//plot.setVisible(true);
					double[] x_ = new double[100];
					double[] b_ = new double[100];
					double[] t_ = new double[100];
					for (int i = 0; i < 100; i++) {
						double curb = 0d;
						double curt = 0d;
						double curx = ((double) i) / 100;
						for (byte j = 0; j < cb.length; j++) curb += cb[j]*Math.pow(curx, j);
						for (byte j = 0; j < ct.length; j++) curt += ct[j]*Math.pow(curx, j);
						x_[i] = curx;
						b_[i] = curb;
						t_[i] = curt;
					}
					
					MyPlot plotb = new MyPlot("Base pressure", 2, "Pressure [hPa]", x, b, x_, b_);
					plotb.setVisible(true);
					
					MyPlot plott = new MyPlot("Base temperature", 2, "Temperature [K]", x, t, x_, t_);
					plott.setVisible(true);	

				} // end of GRAL plotting
				
				
				{ // this block calculates and stores the true altitude into all tracklogs
					System.out.println("Calculating and storing into tracklogs the true altitude");
					for (int tr = 0; tr < tracks.size(); tr++) {
						if (! tracks.get(tr).isEmpty()) {
							int s = 0;
							for (int i = 0; i < tracks.get(tr).getFixNum(); i++) {
								if (tracks.get(tr).isComplete()) {
									double curp = tracks.get(tr).getPressure(i);
									double time = ((double)(tracks.get(tr).getTime(i) - mintime))/((double)(maxtime - mintime));
									double curb = 0d;
									double curt = 0d;
									for (byte j = 0; j < cb.length; j++) curb += cb[j]*Math.pow(time, j);
									for (byte k = 0; k < ct.length; k++) curt += ct[k]*Math.pow(time, k);
									int a = (int) Math.round( (curt/LAPSR) * ( 1d - Math.pow(curp/curb, ATMIC)) );
									tracks.get(tr).setTrueAlt(i, a);
									s += a;
								}
							}
							int avgGpsAlt = tracks.get(tr).getAvgGpsAlt() - (tracks.get(tr).getGeoid() ? 0 : geoidH);
							int avgTrueAlt = s / tracks.get(tr).getFixNum();
							tracks.get(tr).setOffset(avgGpsAlt - avgTrueAlt);
						}
					}
					System.out.println("Calculating true altitude done");
				}
				
				{ // this block saves the tracklogs with true altitude
					if (! noSave) {
						DateTimeFormatter fmt = DateTimeFormat.forPattern("yyyyMMdd_HHmmss");
						DateTime dt = new DateTime();
						String dirString = fmt.print(dt);
						File originalPath = path.toFile();
						File newPathF = new File(originalPath, dirString);
						Path newPathP = newPathF.toPath();
						Files.createDirectory(newPathP);
						
						Path fn;
						File newFile;
						for (int i = 0; i < tracks.size(); i++){
							if (tracks.get(i).isComplete()) {
								fn = tracks.get(i).getSource().getFileName();
								newFile = new File(newPathF, fn.toString());
								System.out.println("Saving file " + newFile.toString() + " with offset " + tracks.get(i).getOffset());
								tracks.get(i).save(newFile.toPath(), ! noTrueAltitude, saveExperimental);
							}
						}
						
						System.out.println("Saving done");
					}
				}
				
				{ // block making the analysis of the identical tracklogs
					if (identtr1 != -1 && identtr2 != -1) {
						long sba = 0;
						long qba = 0;
						long sga = 0;
						long qga = 0;
						long sta = 0;
						long qta = 0;
						long se = 0;
						long qe = 0;
						long sn = 0;
						long qn = 0;
						int avba, avga, avta, ave, avn;
						double sdba, sdga, sdta, sde, sdn;
						int mdba = 0;
						int mdga = 0;
						int mdta = 0;
						int mde = 0;
						int mdn = 0;
						int pn = 0;
						int[] diffba = new int[Math.max(tracks.get(identtr1).getFixNum(), tracks.get(identtr2).getFixNum())];
						int[] diffga = new int[Math.max(tracks.get(identtr1).getFixNum(), tracks.get(identtr2).getFixNum())];
						int[] diffta = new int[Math.max(tracks.get(identtr1).getFixNum(), tracks.get(identtr2).getFixNum())];
						int[] diffe = new int[Math.max(tracks.get(identtr1).getFixNum(), tracks.get(identtr2).getFixNum())];
						int[] diffn = new int[Math.max(tracks.get(identtr1).getFixNum(), tracks.get(identtr2).getFixNum())];
						for (int time = mintime; time < maxtime; time += 1) {
							int i1 = tracks.get(identtr1).getIndex(time, 0);
							int i2 = tracks.get(identtr2).getIndex(time, 0);
							if (i1 >= 0 && i2 >= 0) {
								diffba[pn] = tracks.get(identtr1).getBaroAlt(i1) - tracks.get(identtr2).getBaroAlt(i2);
								diffga[pn] = tracks.get(identtr1).getGpsAlt(i1) - tracks.get(identtr2).getGpsAlt(i2);
								diffta[pn] = tracks.get(identtr1).getTrueAlt(i1) + tracks.get(identtr1).getOffset() 
											- tracks.get(identtr2).getTrueAlt(i2) - tracks.get(identtr2).getOffset();
								int e1 = tracks.get(identtr1).getEastning(i1);
								int n1 = tracks.get(identtr1).getNorthning(i1);
								int e2 = tracks.get(identtr2).getEastning(i2);
								int n2 = tracks.get(identtr2).getNorthning(i2);
								diffe[pn] = e1 - e2; //(int) Math.sqrt( Math.pow((double) (e1 - e2), 2d) + Math.pow((double) (n1 - n2), 2d) );
								diffn[pn] = n1 - n2;
								sba += diffba[pn];
								sga += diffga[pn];
								sta += diffta[pn];
								se += diffe[pn];
								sn += diffn[pn];
								pn++;
							}
						}
						avba = (int) (sba / pn);
						avga = (int) (sga / pn);
						avta = (int) (sta / pn);
						ave = (int) (se / pn);
						avn = (int) (sn / pn);
						for (int i = 0; i < pn; i++) {
							qba += (diffba[i] - avba) * (diffba[i] - avba);
							qga += (diffga[i] - avga) * (diffga[i] - avga);
							qta += (diffta[i] - avta) * (diffta[i] - avta);
							qe += (diffe[i] - ave) * (diffe[i] - ave);
							qn += (diffn[i] - avn) * (diffn[i] - avn);
							if (Math.abs(diffba[i]) > mdba) mdba = Math.abs(diffba[i]);
							if (Math.abs(diffga[i]) > mdga) mdga = Math.abs(diffga[i]);
							if (Math.abs(diffta[i]) > mdta) mdta = Math.abs(diffta[i]);
							if (Math.abs(diffe[i]) > mde) mde = Math.abs(diffe[i]);
							if (Math.abs(diffn[i]) > mdn) mdn = Math.abs(diffn[i]);
						}
						sdba = Math.sqrt(qba / (pn - 1));
						sdga = Math.sqrt(qga / (pn - 1));
						sdta = Math.sqrt(qta / (pn - 1));
						sde = Math.sqrt(qe / (pn - 1));
						sdn = Math.sqrt(qn / (pn - 1));
						System.out.println("Baro alt diff Average " + avba + "  Max " + mdba + "  StDev " + sdba);
						System.out.println("GPS  alt diff Average " + avga + "  Max " + mdga + "  StDev " + sdga);
						System.out.println("True alt diff Average " + avta + "  Max " + mdta + "  StDev " + sdta);
						System.out.println("Eastning diff Average " + ave + "  Max " + mde + "  StDev " + sde);
						System.out.println("Northning diff Average " + avn + "  Max " + mdn + "  StDev " + sdn);
						
						
						/* correlation
						for (i=0; i<pn; i++) {
							arr1[i] -= mean1;
							arr2[i][j] -= mean2;
							mag1 += arr1[i][j] * arr1[i][j];
							mag2 += arr2[i][j] * arr2[i][j];
							corr += arr1[i][j] * arr2[i][j];
						}
						
						corr /= Math.sqrt(mag1*mag2);
						*/
					}
				}
			} else {
				System.out.println("No enough good points! " + n);
			}

		} // block calculating best fit coeficants
				
		
    } // main method

} // TracklogAnalizer class
